/*
Module : MFCSMS.cpp
Purpose: Implementation for MFC wrapper classes for sending SMS messages
Created: PJN / 27-10-2002
History: PJN / 08-11-2002 1. Now uses "\r" consistently as terminator when sending AT commands to the mobile.
                          2. The exception class now uses HRESULT based error values instead of SDK
                          style GetLastError error codes. This allows user defined error codes to be more
                          easily defined
         PJN / 12-03-2005 1. Addition of a CSMSSerialPort::GetLastResponse method. This last response is 
                          now also included in CSMSExceptions when necessary. Thanks to Serhiy Pavlov
                          for suggesting this addition.
                          2. Write timeouts are now also applied in Set0ReadTimeout which is called by 
                          CSMSSerialPort::Open.
         PJN / 26-03-2005 1. Now calls SetDCB upon opening the port to ensure it is put in a valid state
                          and baud rate. This helps fix problems with some mobile phone models. Thanks to 
                          Serhiy Pavlov for reporting this problem. 
                          2. Open method now allows an explicit baud rate to be specified.
                          3. General tidy up of the exception throwing code
         PJN / 08-07-2006 1. Updated copyright details.
                          2. Code now uses newer C++ style casts instead of C style casts.
                          3. Optimized CSMSMessage constructor code
                          4. Replaced calls to ZeroMemory with memset
                          5. Optimized CSMSSerialPort constructor code
                          6. CSMSSerialPort is no longer derived from CObject as it was not really required
                          7. Code now uses SEVERITY_ERROR define for MAKE_HRESULT calls.
                          8. Renamed AfxThrowSMSException to ThrowSMSException and made it part of the 
                          CSMSSerialPort class.
                          9. Updated the code to clean compile on VC 2005
                          10. Removed EnumSerialPorts module from download
                          11. Updated sample app to warn user when it cannot find any serial ports on start-up
         PJN / 02-08-2008 1. Updated copyright details.
                          2. Removed VC 6 style classwizard comments from the sample app.
                          3. Updated the sample app to clean compile on VC 2008
                          4. Sample app now uses CEnumerateSerial::UsingSetupAPI1 to enumerate serial ports
                          5. Sample app now displays the friendly name of the port also
                          6. Updated code to compile correctly using _ATL_CSTRING_EXPLICIT_CONSTRUCTORS define
                          7. Reworked the exception class to be in line with the other exception classes of
                          the author.
                          8. Code now compiles cleanly using Code Analysis (/analyze)
                          9. The code now only supports VC 2005 or later. 
                          10. Internally the code now opens the serial port in overlapped mode
                          11. Removed the Attach and Detach methods.
                          12. Increased the default timeout to 15 seconds.
                          13. Added support for sending using the UCS2 alphabet. Thanks to Andrew Komiagin
                          for this very nice addition.
         PJN / 10-07-2016 1. Updated copyright details.
                          2. Updated the project settings to more modern defaults.
                          3. Added SAL annotations to all the code
                          4. Updated the documentation to reference the specs which MfcSMS uses.
                          5. Reworked the classes to optionally compile without MFC. By default the classes 
                          now use STL classes and idioms but if you define CMFCSMS_MFC_EXTENSIONS the classes 
                          will revert back to the MFC behaviour.
                          6. The serial port is now left in the default OS initialized state on opening. In 
                          the process the baud rate parameter to CSMSSerialPort::Open & 
                          CSMSSerialPort::InitializePort has now been removed.
                          7. Reimplemented the 7bit GSM encoding into easier to understand helper methods
                          8. Implemented a rescan button in the sample app
                          9. Reworked the WaitForSerialResponse method to time out correctly on bad responses
                          10. Removed the restriction on sending a 0 length message. In the process the
                          E_SMS_NO_MESSAGE_TO_PDU_ENCODE error code has been removed.
                          11. Updated the sample app to support an automatic mode to auto detect the Service 
                          Center Number address type and the Phone Number address type.
                          12. Added support for sending the message as raw 8bit encoding.
                          13. Added support for sending concatenated SMS messages using UDHs.
                          14. Updated the sample app to display detailed statistics about the text message
                          which will be sent as the user types the message. This new functionality is 
                          available via the new CSMSMessage::Statistics method.
                          15. Updated the code to clean compile on VC 2008 - VC 2015
         PJN / 20-05-2017 1. Updated copyright details.
                          2. Updated the code to compile cleanly using /permissive-.
         PJN / 24-09-2017 1. Replaced NULL throughout the codebase with nullptr. This means that the minimum
                          requirement for the framework is now VC 2010.
                          2. Replaced CString::operator LPC*STR() calls throughout the codebase with
                          CString::GetString calls
         PJN / 19-11-2017 1. Updated the code to compile cleanly when _ATL_NO_AUTOMATIC_NAMESPACE is defined.
         PJN / 21-10-2018 1. Updated copyright details.
                          2. Fixed a number of C++ core guidelines compiler warnings. These changes mean that 
                          the code will now only compile on VC 2017 or later.
                          3. Removed all code which supported CMFCSMS_MFC_EXTENSIONS code path
         PJN / 21-09-2019 1. Replaced enum with enum class throughout the code
                          2. Updated copyright details
         PJN / 24-12-2019 1. Fixed various Clang-Tidy static code analysis warnings in the code.
                          2. Replaced BOOL with bool throughout the code.
         PJN / 29-03-2020 1. Updated copyright details.
                          2. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 09-05-2020 1. Fixed more Clang-Tidy static code analysis warnings in the code.
         PJN / 02-05-2022 1. Updated copyright details.
                          2. Updated the code to use C++ uniform initialization for all variable declarations.

Copyright (c) 2002 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code.

*/


//////////////////// Includes /////////////////////////////////////////////////

#include "stdafx.h"
#include "MFCSMS.h"
#include "resource.h"
#ifndef _INC_STDLIB
#pragma message("To avoid this message, please put stdlib.h in your pre compiled header (normally stdafx.h)")
#include <stdlib.h>
#endif //#ifndef _INC_STDLIB
#ifndef __ATLSTR_H__
#pragma message("To avoid this message, please put atlstr.h in your pre compiled header (normally stdafx.h)")
#include <atlstr.h>
#endif //#ifndef __ATLSTR_H__


//////////////////// Macros / Defines /////////////////////////////////////////

#ifdef _DEBUG
#define new DEBUG_NEW
#endif //#ifdef _DEBUG

constexpr std::array<WORD, 256> g_smsLookupTable
{ {
  0xFFFF, //Ascii 0
  0xFFFF, //Ascii 1
  0xFFFF, //Ascii 2
  0xFFFF, //Ascii 3
  0xFFFF, //Ascii 4
  0xFFFF, //Ascii 5
  0xFFFF, //Ascii 6
  0xFFFF, //Ascii 7
  0xFFFF, //Ascii 8
  0xFFFF, //Ascii 9
  0x000A, //Ascii 10 "LINE FEED"
  0xFFFF, //Ascii 11
  0x1B0A, //Ascii 12 "FORM FEED"
  0x000D, //Ascii 13 "CARRIAGE RETURN"
  0xFFFF, //Ascii 14
  0xFFFF, //Ascii 15
  0xFFFF, //Ascii 16
  0xFFFF, //Ascii 17
  0xFFFF, //Ascii 18
  0xFFFF, //Ascii 19
  0xFFFF, //Ascii 20
  0xFFFF, //Ascii 21
  0xFFFF, //Ascii 22
  0xFFFF, //Ascii 23
  0xFFFF, //Ascii 24
  0xFFFF, //Ascii 25
  0xFFFF, //Ascii 26
  0xFFFF, //Ascii 27
  0xFFFF, //Ascii 28
  0xFFFF, //Ascii 29
  0xFFFF, //Ascii 30
  0xFFFF, //Ascii 31
  0x0020, //Ascii 32 " "
  0x0021, //Ascii 33 "!"
  0x0022, //Ascii 34 """
  0x0023, //Ascii 35 "#"
  0x0002, //Ascii 36 "$"
  0x0025, //Ascii 37 "%"
  0x0026, //Ascii 38 "&"
  0x0027, //Ascii 39 "'"
  0x0028, //Ascii 40 "("
  0x0029, //Ascii 41 ")"
  0x002A, //Ascii 42 "*"
  0x002B, //Ascii 43 "+"
  0x002C, //Ascii 44 ","
  0x002D, //Ascii 45 "-"
  0x002E, //Ascii 46 "."
  0x002F, //Ascii 47 "/"
  0x0030, //Ascii 48 "0"
  0x0031, //Ascii 49 "1"
  0x0032, //Ascii 50 "2"
  0x0033, //Ascii 51 "3"
  0x0034, //Ascii 52 "4"
  0x0035, //Ascii 53 "5"
  0x0036, //Ascii 54 "6"
  0x0037, //Ascii 55 "7"
  0x0038, //Ascii 56 "8"
  0x0039, //Ascii 57 "9"
  0x003A, //Ascii 58 ":"
  0x003B, //Ascii 59 ";"
  0x003C, //Ascii 60 "<"
  0x003D, //Ascii 61 "="
  0x003E, //Ascii 62 ">"
  0x003F, //Ascii 63 "?"
  0x0000, //Ascii 64 "@"
  0x0041, //Ascii 65 "A"
  0x0042, //Ascii 66 "B"
  0x0043, //Ascii 67 "C"
  0x0044, //Ascii 68 "D"
  0x0045, //Ascii 69 "E"
  0x0046, //Ascii 70 "F"
  0x0047, //Ascii 71 "G"
  0x0048, //Ascii 72 "H"
  0x0049, //Ascii 73 "I"
  0x004A, //Ascii 74 "J"
  0x004B, //Ascii 75 "K"
  0x004C, //Ascii 76 "L"
  0x004D, //Ascii 77 "M"
  0x004E, //Ascii 78 "N"
  0x004F, //Ascii 79 "O"
  0x0050, //Ascii 80 "P"
  0x0051, //Ascii 81 "Q"
  0x0052, //Ascii 82 "R"
  0x0053, //Ascii 83 "S"
  0x0054, //Ascii 84 "T"
  0x0055, //Ascii 85 "U"
  0x0056, //Ascii 86 "V"
  0x0057, //Ascii 87 "W"
  0x0058, //Ascii 88 "X"
  0x0059, //Ascii 89 "Y"
  0x005A, //Ascii 90 "Z"
  0x1B3C, //Ascii 91 "["
  0x1B2F, //Ascii 92 "\"
  0x1B3E, //Ascii 93 "]"
  0x1B14, //Ascii 94 "^"
  0x0011, //Ascii 95 "_"
  0xFFFF, //Ascii 96
  0x0061, //Ascii 97 "a"
  0x0062, //Ascii 98 "b"
  0x0063, //Ascii 99 "c"
  0x0064, //Ascii 100 "d"
  0x0065, //Ascii 101 "e"
  0x0066, //Ascii 102 "f"
  0x0067, //Ascii 103 "g"
  0x0068, //Ascii 104 "h"
  0x0069, //Ascii 105 "i"
  0x006A, //Ascii 106 "j"
  0x006B, //Ascii 107 "k"
  0x006C, //Ascii 108 "l"
  0x006D, //Ascii 109 "m"
  0x006E, //Ascii 110 "n"
  0x006F, //Ascii 111 "o"
  0x0070, //Ascii 112 "p"
  0x0071, //Ascii 113 "q"
  0x0072, //Ascii 114 "r"
  0x0073, //Ascii 115 "s"
  0x0074, //Ascii 116 "t"
  0x0075, //Ascii 117 "u"
  0x0076, //Ascii 118 "v"
  0x0077, //Ascii 119 "w"
  0x0078, //Ascii 120 "x"
  0x0079, //Ascii 121 "y"
  0x007A, //Ascii 122 "z"
  0x1B28, //Ascii 123 "{"
  0x1B40, //Ascii 124 "|"
  0x1B29, //Ascii 125 "}"
  0x1B3D, //Ascii 126 "~"
  0xFFFF, //Ascii 127
  0x1B65, //Ascii 128 "EURO SYMBOL", Updated Latin-1
  0xFFFF, //Ascii 129
  0xFFFF, //Ascii 130
  0xFFFF, //Ascii 131
  0xFFFF, //Ascii 132
  0xFFFF, //Ascii 133
  0xFFFF, //Ascii 134
  0xFFFF, //Ascii 135
  0xFFFF, //Ascii 136
  0xFFFF, //Ascii 137
  0xFFFF, //Ascii 138
  0xFFFF, //Ascii 139
  0xFFFF, //Ascii 140
  0xFFFF, //Ascii 141
  0xFFFF, //Ascii 142
  0xFFFF, //Ascii 143
  0xFFFF, //Ascii 144
  0xFFFF, //Ascii 145
  0xFFFF, //Ascii 146
  0xFFFF, //Ascii 147
  0xFFFF, //Ascii 148
  0xFFFF, //Ascii 149
  0xFFFF, //Ascii 150
  0xFFFF, //Ascii 151
  0xFFFF, //Ascii 152
  0xFFFF, //Ascii 153
  0xFFFF, //Ascii 154
  0xFFFF, //Ascii 155
  0xFFFF, //Ascii 156
  0xFFFF, //Ascii 157
  0xFFFF, //Ascii 158
  0xFFFF, //Ascii 159
  0xFFFF, //Ascii 160
  0x0040, //Ascii 161 ""
  0xFFFF, //Ascii 162
  0x0001, //Ascii 163 ""
  0x0024, //Ascii 164 ""
  0x0003, //Ascii 165 ""
  0xFFFF, //Ascii 166
  0x005F, //Ascii 167 ""
  0xFFFF, //Ascii 168
  0xFFFF, //Ascii 169
  0xFFFF, //Ascii 170
  0xFFFF, //Ascii 171
  0xFFFF, //Ascii 172
  0xFFFF, //Ascii 173
  0xFFFF, //Ascii 174
  0xFFFF, //Ascii 175
  0xFFFF, //Ascii 176
  0xFFFF, //Ascii 177
  0xFFFF, //Ascii 178
  0xFFFF, //Ascii 179
  0xFFFF, //Ascii 180
  0xFFFF, //Ascii 181
  0xFFFF, //Ascii 182
  0xFFFF, //Ascii 183
  0xFFFF, //Ascii 184
  0xFFFF, //Ascii 185
  0xFFFF, //Ascii 186
  0xFFFF, //Ascii 187
  0xFFFF, //Ascii 188
  0xFFFF, //Ascii 189
  0xFFFF, //Ascii 190
  0x0060, //Ascii 191 ""
  0xFFFF, //Ascii 192
  0xFFFF, //Ascii 193
  0xFFFF, //Ascii 194
  0xFFFF, //Ascii 195
  0x005B, //Ascii 196 ""
  0x000E, //Ascii 197 ""
  0x001C, //Ascii 198 ""
  0x0009, //Ascii 199 ""
  0xFFFF, //Ascii 200
  0x001F, //Ascii 201 ""
  0xFFFF, //Ascii 202
  0xFFFF, //Ascii 203
  0xFFFF, //Ascii 204
  0xFFFF, //Ascii 205
  0xFFFF, //Ascii 206
  0xFFFF, //Ascii 207
  0xFFFF, //Ascii 208
  0x005D, //Ascii 209 ""
  0xFFFF, //Ascii 210
  0xFFFF, //Ascii 211
  0xFFFF, //Ascii 212
  0xFFFF, //Ascii 213
  0x005C, //Ascii 214 ""
  0xFFFF, //Ascii 215
  0x000B, //Ascii 216 ""
  0xFFFF, //Ascii 217
  0xFFFF, //Ascii 218
  0xFFFF, //Ascii 219
  0x005E, //Ascii 220 ""
  0xFFFF, //Ascii 221
  0xFFFF, //Ascii 222
  0x001E, //Ascii 223 ""
  0x007F, //Ascii 224 ""
  0xFFFF, //Ascii 225
  0xFFFF, //Ascii 226
  0xFFFF, //Ascii 227
  0x007B, //Ascii 228 ""
  0x000F, //Ascii 229 ""
  0x001D, //Ascii 230 ""
  0xFFFF, //Ascii 231
  0x0004, //Ascii 232 ""
  0x0005, //Ascii 233 ""
  0xFFFF, //Ascii 234
  0xFFFF, //Ascii 235
  0x0007, //Ascii 236 ""
  0xFFFF, //Ascii 237
  0xFFFF, //Ascii 238
  0xFFFF, //Ascii 239
  0xFFFF, //Ascii 240
  0x007D, //Ascii 241 ""
  0x0008, //Ascii 242 ""
  0xFFFF, //Ascii 243
  0xFFFF, //Ascii 244
  0xFFFF, //Ascii 245
  0x007C, //Ascii 246 ""
  0xFFFF, //Ascii 247
  0x000C, //Ascii 248 ""
  0x0006, //Ascii 249 ""
  0xFFFF, //Ascii 250
  0xFFFF, //Ascii 251
  0x007E, //Ascii 252 ""
  0xFFFF, //Ascii 253
  0xFFFF, //Ascii 254
  0xFFFF, //Ascii 255
} };


///////////////////////// Implementation //////////////////////////////////////

CSMSException::CSMSException(_In_ HRESULT hr, _In_ CMFCSMSString sLastResponse) noexcept : m_hr{hr},
                                                                                           m_sLastResponse{std::move(sLastResponse)}
{
}

CSMSException::CSMSException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_ CMFCSMSString sLastResponse) noexcept : m_hr{MAKE_HRESULT(SEVERITY_ERROR, dwFacility, dwError)},
                                                                                                                     m_sLastResponse{std::move(sLastResponse)}
{
}

#pragma warning(suppress: 26429)
BOOL CSMSException::GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(lpszError != nullptr);

  //We don't support help context at the moment
  if (pnHelpContext != nullptr)
    *pnHelpContext = 0;

  //What will be the return value from this function (assume the worst)
  BOOL bSuccess{FALSE};

  if (HRESULT_FACILITY(m_hr) == FACILITY_ITF)
  {
    bSuccess = TRUE;

    //Simply load up the string from the string table
    ATL::CAtlString sMsg;
    if (sMsg.LoadString(GetModuleHandle(nullptr), HRESULT_CODE(m_hr)))
      sMsg.Replace(_T("%1"), m_sLastResponse.c_str());
    ATL::Checked::tcsncpy_s(lpszError, nMaxError, sMsg, _TRUNCATE);
  }
  else
  {
    LPTSTR lpBuffer{nullptr};
#pragma warning(suppress: 26490)
    const DWORD dwReturn{FormatMessage(FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
                                       nullptr, HRESULT_CODE(m_hr), MAKELANGID(LANG_NEUTRAL, SUBLANG_SYS_DEFAULT),
                                       reinterpret_cast<LPTSTR>(&lpBuffer), 0, nullptr)};
    if (dwReturn == 0)
      *lpszError = _T('\0');
    else
    {
      bSuccess = TRUE;
      ATL::Checked::tcsncpy_s(lpszError, nMaxError, lpBuffer, _TRUNCATE);
      LocalFree(lpBuffer);
    }
  }

  return bSuccess;
}


CSMSMessageStats::CSMSMessageStats() noexcept : m_nUserDataSizeCharacters{0},
                                                m_nUserDataSizeBytes{0},
                                                m_nMessages{0},
                                                m_nMaxBytesPerMessage{0},
                                                m_nMaxCharactersPerMessage{0},
                                                m_nExtendedCharacters{0},
                                                m_nReplacementCharacters{0}
{
}


CSMSMessage::CSMSMessage() noexcept : m_ServiceCenterNumberAddressType{static_cast<BYTE>(PhoneFormat::UnknownFormat)},
                                      m_PhoneNumberAddressType{static_cast<BYTE>(PhoneFormat::UnknownFormat)},
                                      m_bUseValidityPeriod{false},
                                      m_bUseRelativeValidityPeriod{true},
                                      m_RelativeValidityPeriod{1, 0, 0, 0},
                                      m_AbsoluteValidityPeriod{},
                                      m_TimeZone{0},
                                      m_bRequireStatusReport{false},
                                      m_cReplacementCharacter{_T('_')},
                                      m_Alphabet{Alphabet::GSM_7BIT}
{
#pragma warning(suppress: 26472)
  m_byReferenceNumber = static_cast<BYTE>(rand() % 256); //Use a random reference number by default
}

CSMSMessageStats CSMSMessage::Statistics()
{
  //What will be the return value from this method
  CSMSMessageStats stats;

  //First thing to do is work out the full size of the user data we are sending
  std::vector<BYTE> userData;
  switch (m_Alphabet)
  {
    case CSMSMessage::Alphabet::GSM_7BIT:
    {
      //Convert the message to GSM 7 bit
      ConvertToGSM7Bit(m_sMessage, m_cReplacementCharacter, userData, &stats.m_nExtendedCharacters, &stats.m_nReplacementCharacters);
      break;
    }
    case CSMSMessage::Alphabet::UCS2:
    {
      //Convert the message to UCS2
      ConvertToUCS2(m_sMessage, userData);
      break;
    }
    case CSMSMessage::Alphabet::EIGHTBIT:
    {
      userData.insert(userData.end(), m_byMessage.begin(), m_byMessage.end());
      break;
    }
    default:
    {
#pragma warning(suppress: 26477)
      ATLASSERT(FALSE);
      break;
    }
  }
#pragma warning(suppress: 26472)
  stats.m_nUserDataSizeBytes = static_cast<int>(userData.size());

  switch (m_Alphabet)
  {
    case CSMSMessage::Alphabet::GSM_7BIT:
    {
      if (stats.m_nUserDataSizeBytes > 160)
      {
        stats.m_nMessages = (stats.m_nUserDataSizeBytes + 152) / 153;
        stats.m_nMaxCharactersPerMessage = 153;
        stats.m_nMaxBytesPerMessage = 134;
      }
      else
      {
        stats.m_nMessages = 1;
        stats.m_nMaxCharactersPerMessage = 160;
        stats.m_nMaxBytesPerMessage = 140;
      }
#pragma warning(suppress: 26472)
      stats.m_nUserDataSizeCharacters = static_cast<int>(m_sMessage.length());
      break;
    }
    case CSMSMessage::Alphabet::UCS2:
    {
      if (stats.m_nUserDataSizeBytes > 140)
      {
        stats.m_nMessages = (stats.m_nUserDataSizeBytes + 133) / 134;
        stats.m_nMaxCharactersPerMessage = 67;
        stats.m_nMaxBytesPerMessage = 134;
      }
      else
      {
        stats.m_nMessages = 1;
        stats.m_nMaxCharactersPerMessage = 70;
        stats.m_nMaxBytesPerMessage = 140;
      }
#pragma warning(suppress: 26472)
      stats.m_nUserDataSizeCharacters = static_cast<int>(m_sMessage.length());
      break;
    }
    case CSMSMessage::Alphabet::EIGHTBIT:
    {
      if (stats.m_nUserDataSizeBytes > 140)
      {
        stats.m_nMessages = (stats.m_nUserDataSizeBytes + 133) / 134;
        //We leave stats.m_nMaxCharactersPerMessage == 0 for the Eight bit alphabet
        stats.m_nMaxBytesPerMessage = 134;
      }
      else
      {
        stats.m_nMessages = 70;
        //We leave stats.m_nMaxCharactersPerMessage == 0 for the Eight bit alphabet
        stats.m_nMessages = 1;
        stats.m_nMaxBytesPerMessage = 140;
      }
      //We leave stats.m_nUserDataSizeCharacters == 0 for the Eight bit alphabet
      break;
    }
    default:
    {
#pragma warning(suppress: 26477)
      ATLASSERT(FALSE);
      break;
    }
  }

  return stats;
}

void CSMSMessage::ConvertToUCS2(_In_ const CMFCSMSString& sData, _Inout_ std::vector<BYTE>& EncodedData)
{
  //Convert the input string to UCS2
  ATL::CT2W sWideString{sData.c_str()};
  const size_t nConvertedChars{wcslen(sWideString.operator LPWSTR())};

  //Convert the UCS2 string to a byte array
  EncodedData.resize(2 * nConvertedChars);
  size_t nOutputOffset{0};
  for (size_t i{0}; i<nConvertedChars; i++)
  {
#pragma warning(suppress: 26446 26472 26481)
    EncodedData[nOutputOffset] = static_cast<BYTE>((sWideString[i] & 0xFF00) >> 8); //NOLINT(clang-analyzer-core.UndefinedBinaryOperatorResult)
    ++nOutputOffset;
#pragma warning(suppress: 26446 26472 26481)
    EncodedData[nOutputOffset] = static_cast<BYTE>(sWideString[i] & 0xFF);
    ++nOutputOffset;
  }
}

void CSMSMessage::ConvertToGSM7Bit(_In_ const CMFCSMSString& sData, _In_ TCHAR cReplacementCharacter, _Inout_ std::vector<BYTE>& EncodedData, _Out_opt_ int* pnExtendedCharacters, _Out_opt_ int* pnReplacementCharacters)
{
#pragma warning(suppress: 26472)
  const int nDataLength{static_cast<int>(sData.length())};
  EncodedData.clear();

  //Set the output parameters to a default if required
  if (pnExtendedCharacters != nullptr)
    *pnExtendedCharacters = 0;
  if (pnReplacementCharacters != nullptr)
    *pnReplacementCharacters = 0;

  //Get the mapping for the replacement character
#pragma warning(suppress: 26446 26482)
  const WORD wReplacementCharacter{g_smsLookupTable[TCHAR2BYTE(cReplacementCharacter)]};
#pragma warning(suppress: 26477)
  ATLASSERT(wReplacementCharacter != 0xFFFF); //Replacement character HAS to map to something!

  //Work through all the data
  for (int i{0}; i<nDataLength; i++)
  {
#pragma warning(suppress: 26446 26482)
    const WORD wCurrentChar{g_smsLookupTable[TCHAR2BYTE(sData[i])]};
    if (wCurrentChar == 0xFFFF)
    {
      //Update the output parameter if required
      if (pnReplacementCharacters != nullptr)
        (*pnReplacementCharacters)++;

      if (wReplacementCharacter > 0x7F)
      {
        //Update the output parameter if required
        if (pnExtendedCharacters != nullptr)
          (*pnExtendedCharacters)++;

#pragma warning(suppress: 26489)
        EncodedData.push_back(static_cast<BYTE>((wReplacementCharacter & 0xFF00) >> 8));
#pragma warning(suppress: 26489)
        EncodedData.push_back(static_cast<BYTE>(wReplacementCharacter & 0x00FF));
      }
      else
#pragma warning(suppress: 26489)
        EncodedData.push_back(static_cast<BYTE>(wReplacementCharacter));
    }
    else if (wCurrentChar > 0x7F)
    {
      //Update the output parameter if required
      if (pnExtendedCharacters != nullptr)
        (*pnExtendedCharacters)++;

#pragma warning(suppress: 26489)
      EncodedData.push_back(static_cast<BYTE>((wCurrentChar & 0xFF00) >> 8));
#pragma warning(suppress: 26489)
      EncodedData.push_back(static_cast<BYTE>(wCurrentChar & 0x00FF));
    }
    else
#pragma warning(suppress: 26489)
      EncodedData.push_back(static_cast<BYTE>(wCurrentChar));
  }
}

BYTE CSMSMessage::TCHAR2BYTE(_In_ TCHAR cCharacter)
{
  std::array<TCHAR, 2> szData{{cCharacter, _T('\0')}};
#pragma warning(suppress: 26485)
  ATL::CT2A szAsciiData{szData.data()};
#pragma warning(suppress: 26481)
  return szAsciiData[0];
}


CSMSSerialPort::CSMSSerialPort() noexcept : m_hComm{INVALID_HANDLE_VALUE},
                                            m_hEvent{nullptr},
                                            m_dwTimeout{15000},
                                            m_Overlapped{}
{
}

CSMSSerialPort::~CSMSSerialPort()
{
  Close();
}

void CSMSSerialPort::ThrowSMSException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_ const CMFCSMSString& sLastResponse)
{
  if (dwError == 0)
    dwError = GetLastError();

  ATLTRACE(_T("Warning: throwing CSMSException for error %08X\n"), dwError);
  CSMSException e{dwError, dwFacility, sLastResponse};
  throw e;
}

void CSMSSerialPort::ThrowSMSException(_In_ HRESULT hr, _In_ const CMFCSMSString& sLastResponse)
{
  ATLTRACE(_T("Warning: throwing CSMSException for error %08X\n"), hr);
  CSMSException e{hr, sLastResponse};
  throw e;
}

void CSMSSerialPort::Open(_In_ int nPort, _In_ DWORD dwTimeout)
{
  Close(); //In case we are already open

  //Call CreateFile to open up the comms port
  ATL::CAtlString sPort;
  sPort.Format(_T("\\\\.\\COM%d"), nPort);
  m_hComm = CreateFile(sPort, GENERIC_READ | GENERIC_WRITE, 0, nullptr, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, nullptr);
  if (m_hComm == INVALID_HANDLE_VALUE)
    ThrowSMSException(GetLastError(), FACILITY_WIN32);

  //Create the event we need for later synchronisation use
  m_hEvent = CreateEvent(nullptr, TRUE, FALSE, nullptr);
  if (m_hEvent == nullptr)
  {
    const DWORD dwError{GetLastError()};

    //Close the port before we throw the exception
    Close();

    ThrowSMSException(dwError, FACILITY_WIN32);
  }
  m_Overlapped.hEvent = m_hEvent;

  //Set the timeouts on the port
  InitializePort(dwTimeout);
}

void CSMSSerialPort::Close() noexcept
{
  if (IsOpen())
  {
    //Close down the comms port
    CloseHandle(m_hComm);
    m_hComm = INVALID_HANDLE_VALUE;
  }
  
  //Free the event object we are using
  if (m_hEvent)
  {
    CloseHandle(m_hEvent);
    m_hEvent = nullptr;
  }
}

DWORD CSMSSerialPort::Read(_Out_writes_bytes_(dwCount) __out_data_source(FILE) void* lpBuf, _In_ DWORD dwCount)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwBytesRead{0};
  const BOOL bSuccess{ReadFile(m_hComm, lpBuf, dwCount, &dwBytesRead, &m_Overlapped)};
  if (!bSuccess && (GetLastError() == ERROR_IO_PENDING))
  {
    //Wait for the result if the operation has to be completed in overlapped mode
    GetOverlappedResult(TRUE, dwBytesRead);
  }
  else if (!bSuccess)
    ThrowSMSException(GetLastError(), FACILITY_WIN32);

  return dwBytesRead;
}

DWORD CSMSSerialPort::Write(_In_reads_bytes_opt_(dwCount) const void* lpBuf, _In_ DWORD dwCount)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwBytesWritten{0};
  const BOOL bSuccess{WriteFile(m_hComm, lpBuf, dwCount, &dwBytesWritten, &m_Overlapped)};
  if (!bSuccess && (GetLastError() == ERROR_IO_PENDING))
  {
    //Wait for the result if the operation has to be completed in overlapped mode
    GetOverlappedResult(TRUE, dwBytesWritten);
  }
  else if (!bSuccess)
    ThrowSMSException(GetLastError(), FACILITY_WIN32);

  return dwBytesWritten;
}

bool CSMSSerialPort::DataWaiting(_In_ DWORD dwTimeout)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  DWORD dwEvtMask{0};
  BOOL bSuccess{WaitCommEvent(m_hComm, &dwEvtMask, &m_Overlapped)};
  if (!bSuccess)
  {
    //See if the event gets signalled before the timeout value
    if (WaitForSingleObject(m_hEvent, dwTimeout) == WAIT_OBJECT_0)
    {
      DWORD dwBytesTransferred{0};
      GetOverlappedResult(FALSE, dwBytesTransferred);
      bSuccess = TRUE;
    }
    else
    {
      //Ensure we cancel the pending IO
      CancelIo(m_hComm);
    }
  }

  return bSuccess;
}

void CSMSSerialPort::GetOverlappedResult(_In_ bool bWait, _Out_ DWORD& dwBytesTransferred)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  if (!::GetOverlappedResult(m_hComm, &m_Overlapped, &dwBytesTransferred, bWait))
    ThrowSMSException(GetLastError(), FACILITY_WIN32);
}

void CSMSSerialPort::InitializePort(_In_ DWORD dwTimeout)
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  //Set the comm read timeout to return immediately
  COMMTIMEOUTS Timeouts{};
  Timeouts.ReadIntervalTimeout = MAXDWORD;
  Timeouts.WriteTotalTimeoutConstant = dwTimeout;
  if (!SetCommTimeouts(m_hComm, &Timeouts))
  {
    const DWORD dwError{GetLastError()};

    //Close the port before we throw the exception
    Close();

    ThrowSMSException(dwError, FACILITY_WIN32);
  }

  //Also setup notification for receiving data from the port
  if (!SetCommMask(m_hComm, EV_RXCHAR))
  {
    const DWORD dwError{GetLastError()};

    //Close the port before we throw the exception
    Close();

    ThrowSMSException(dwError, FACILITY_WIN32);
  }

  //Finally also hive away the timeouts value
  m_dwTimeout = dwTimeout;
}

bool CSMSSerialPort::WaitForSerialResponse()
{
  //Validate our parameters
#pragma warning(suppress: 26477)
  ATLASSERT(IsOpen());

  //What will be the return value from this method (assume the worst)
  bool bSuccess{false};

  //retrieve the response until we get the terminator or a timeout occurs
  std::string sResponse;
  m_sLastResponse.clear();
  bool bFoundTerminator{false};
  while (!bFoundTerminator)
  {
    //Has the timeout occurred
    if (!DataWaiting(m_dwTimeout))
    {
      //Hive away the last response
      m_sLastResponse = ATL::CA2T{sResponse.c_str()};
      return false;
    }

    //receive the data from the serial port
    ATL::CAtlStringA sBuf;
#pragma warning(suppress: 26429)
    LPSTR pszBuf{sBuf.GetBuffer(4096)};
    DWORD dwData{Read(pszBuf, 4095)};
    if (dwData)
    {
      //Replace all embedded nulls with spaces
      for (DWORD i{0}; i<dwData; i++)
      {
#pragma warning(suppress: 26481)
        if (pszBuf[i] == '\0')
          pszBuf[i] = ' ';
      }

      //Null terminate the data received
#pragma warning(suppress: 26481)
      pszBuf[dwData] = '\0';

      //Grow the response data
      sResponse += sBuf;

      //Check to see if the terminator character(s) have been found
      bFoundTerminator = (sResponse.find("OK") != std::string::npos);
      if (bFoundTerminator)
        bSuccess = true;
      else
        bFoundTerminator = (sResponse.find("ERROR") != std::string::npos);
    }
    sBuf.ReleaseBuffer();
  }

  //Hive away the last response
  m_sLastResponse = ATL::CA2T{sResponse.c_str()};

  return bSuccess;
}

void CSMSSerialPort::Send(_In_ const CSMSMessage& message)
{
  //Check to see if there is a modem on the specified serial port
  static const char* pszATMsg{"ATV1\r"};
#pragma warning(suppress: 26472)
  Write(pszATMsg, static_cast<DWORD>(strlen(pszATMsg)));
  if (!WaitForSerialResponse())
    ThrowSMSException(E_SMS_NO_MODEM_RESPONSE, m_sLastResponse);

  //Select Message service
  static const char* pszSelectMessageMsg{"AT+CSMS=0\r"};
#pragma warning(suppress: 26472)
  Write(pszSelectMessageMsg, static_cast<DWORD>(strlen(pszSelectMessageMsg)));
  if (!WaitForSerialResponse())
    ThrowSMSException(E_SMS_NO_SMS_SUPPORT, m_sLastResponse);

  //Set the transfer mode to PDU
  static const char* pszMessageFormat{"AT+CMGF=0\r"};
#pragma warning(suppress: 26472)
  Write(pszMessageFormat, static_cast<DWORD>(strlen(pszMessageFormat)));
  if (!WaitForSerialResponse())
    ThrowSMSException(E_SMS_SET_PDU_MODE, m_sLastResponse);

  //First thing to do is work out the full size of the user data we are sending
  std::vector<BYTE> userData;
  switch (message.m_Alphabet)
  {
    case CSMSMessage::Alphabet::GSM_7BIT:
    {
      //Convert the message to GSM 7 bit
      CSMSMessage::ConvertToGSM7Bit(message.m_sMessage, message.m_cReplacementCharacter, userData, nullptr, nullptr);
      break;
    }
    case CSMSMessage::Alphabet::UCS2:
    {
      //Convert the message to UCS2
      CSMSMessage::ConvertToUCS2(message.m_sMessage, userData);
      break;
    }
    case CSMSMessage::Alphabet::EIGHTBIT:
    {
      userData.insert(userData.end(), message.m_byMessage.begin(), message.m_byMessage.end());
      break;
    }
    default:
    {
#pragma warning(suppress: 26477)
      ATLASSERT(FALSE);
      break;
    }
  }
#pragma warning(suppress: 26472)
  int nMessageLength{static_cast<int>(userData.size())};
#pragma warning(suppress: 26446)
  const BYTE* pbyUserData{userData.size() ? userData.data() : nullptr};

  //Next from the size of the userData, work out how many actual SMS messages we will be sending
  int nSMSMessages{0};
  int nSMSMaxLength{0};
  switch (message.m_Alphabet)
  {
    case CSMSMessage::Alphabet::GSM_7BIT:
    {
      nSMSMaxLength = 160;
      if (nMessageLength > nSMSMaxLength)
      {
        nSMSMessages = (nMessageLength + 152) / 153;
        nSMSMaxLength = 153;
      }
      else
        nSMSMessages = 1;
      break;
    }
    case CSMSMessage::Alphabet::UCS2: //Deliberate fallthrough
    case CSMSMessage::Alphabet::EIGHTBIT:
    {
      nSMSMaxLength = 140;
      if (nMessageLength > nSMSMaxLength)
      {
        nSMSMessages = (nMessageLength + 133) / 134;
        nSMSMaxLength = 134;
      }
      else
        nSMSMessages = 1;
      break;
    }
    default:
    {
#pragma warning(suppress: 26477)
      ATLASSERT(FALSE);
      break;
    }
  }

  //The absolute maximum size of message we can encode is what corresponds to 255 concatenated messages
  if (nSMSMessages > 255)
    ThrowSMSException(E_SMS_MESSAGE_TOO_LARGE_TO_PDU_ENCODE);

  //Loop around on each SMS message
  for (int i{0}; i<nSMSMessages; i++)
  {
#pragma warning(suppress: 26489)
    std::vector<BYTE> SMS;

    //First thing to do is work on the Service Center Number
    AppendEncodedSMSCNumber(message.m_sServiceCenterNumber, message.m_ServiceCenterNumberAddressType, SMS);

    //Form the SMS-SUBMIT octet
    BYTE bySMSSubmit{SMS_SUBMIT};
    if (message.m_bUseValidityPeriod)
    {
      if (message.m_bUseRelativeValidityPeriod)
        bySMSSubmit |= UseRelativeValidityPeriod;
      else
        bySMSSubmit |= DoNoUseRelativeValidityPeriod;
    }
    if (message.m_bRequireStatusReport)
      bySMSSubmit |= RequireStatusReport;
    if (nSMSMessages > 1)
      bySMSSubmit |= UseUDH;
    SMS.push_back(bySMSSubmit);
    SMS.push_back(Empty_TP_Message_Reference); //Next is the TP-Message-Reference. We let the phone set the message reference itself

    //Next is the Phone number
#pragma warning(suppress: 26477)
    ATLASSERT(message.m_sPhoneNumber.length()); //Must specify a number to deliver the message to
    AppendEncodedPhoneNumber(message.m_sPhoneNumber, message.m_PhoneNumberAddressType, SMS);

    //Next is the Protocol Identifier
    SMS.push_back(SMS_Type0_Protocol_Identifier);

    //Next is the Data encoding scheme value
    BYTE byCoding{0};
    switch (message.m_Alphabet)
    {
      case CSMSMessage::Alphabet::GSM_7BIT:
      {
        byCoding = Data_Coding_GSM_7Bit;
        break;
      }
      case CSMSMessage::Alphabet::UCS2:
      {
        byCoding = Data_Coding_UCS2;
        break;
      }
      case CSMSMessage::Alphabet::EIGHTBIT:
      {
        byCoding = Data_Coding_8Bit;
        break;
      }
      default:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(FALSE);
        break;
      }
    }
    SMS.push_back(byCoding);

    //Next is the validity period
    if (message.m_bUseValidityPeriod)
    {
      if (message.m_bUseRelativeValidityPeriod)
      {
        //Relative Validity period is pseudo logarithm encoded into 1 octet
        BYTE byValidityPeriod{0};
        if (message.m_RelativeValidityPeriod <= ATL::CTimeSpan{0, 12, 0, 0})
#pragma warning(suppress: 26472)
          byValidityPeriod = static_cast<BYTE>((message.m_RelativeValidityPeriod.GetTotalSeconds() / 300) - 1);
        else if (message.m_RelativeValidityPeriod <= ATL::CTimeSpan{1, 0, 0, 0})
#pragma warning(suppress: 26472)
          byValidityPeriod = static_cast<BYTE>(((message.m_RelativeValidityPeriod.GetTotalSeconds() - 43200) / 1800) + 143);
        else if (message.m_RelativeValidityPeriod <= ATL::CTimeSpan{30, 0, 0, 0})
#pragma warning(suppress: 26472)
          byValidityPeriod = static_cast<BYTE>((message.m_RelativeValidityPeriod.GetTotalSeconds() / 86400) + 166);
        else
#pragma warning(suppress: 26472)
          byValidityPeriod = static_cast<BYTE>((message.m_RelativeValidityPeriod.GetTotalSeconds() / 604800) + 192);

        SMS.push_back(byValidityPeriod);
      }
      else
      {
        //Absolute Validity period is encoded into 7 octets
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wYear % 100));
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wMonth));
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wDay));
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wHour));
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wMinute));
        SMS.push_back(SwappedNibble(message.m_AbsoluteValidityPeriod.wSecond));

        const long nLocalTimeZone{labs(message.m_TimeZone)};
        BYTE byTimeZone{SwappedNibble(nLocalTimeZone / 15)};
        if (message.m_TimeZone < 0)
          byTimeZone |= 0x80;
        SMS.push_back(byTimeZone);
      }
    }

    //Initially we just add a place holder as it will be updated later on once we have added the actual data
#pragma warning(suppress: 26472)
    const int nDataLengthIndex{static_cast<int>(SMS.size())};
    SMS.push_back(0);

    //Handle the User Data header if required
    if (nSMSMessages > 1)
    {
      SMS.push_back(5); //UDHL
      SMS.push_back(IEI_ConcatenatedSMS); //IEI
      SMS.push_back(3); //IEDL
      SMS.push_back(message.m_byReferenceNumber);
      SMS.push_back(static_cast<BYTE>(nSMSMessages));
      SMS.push_back(static_cast<BYTE>(i + 1));
    }

    //Handle encoding to GSM 7bit, UCS2 or 8bit
    switch (message.m_Alphabet)
    {
      case CSMSMessage::Alphabet::GSM_7BIT:
      {
        int nThisMessageLength{min(nMessageLength, nSMSMaxLength)};
        AppendGSM7Bit(pbyUserData, nThisMessageLength, nSMSMessages > 1 ? 1 : 0, SMS);

        //Update the data length now that we know the exact size of it (include the UDH length in septets also if necessary)
        if (nSMSMessages > 1)
          nThisMessageLength += 7;
#pragma warning(suppress: 26446 26472)
        SMS[nDataLengthIndex] = static_cast<BYTE>(nThisMessageLength);
        break;
      }
      case CSMSMessage::Alphabet::UCS2: //deliberate fallthrough
      case CSMSMessage::Alphabet::EIGHTBIT:
      {
        int nThisMessageLength{min(nMessageLength, nSMSMaxLength)};
#pragma warning(suppress: 26481)
        SMS.insert(SMS.end(), pbyUserData, pbyUserData + nThisMessageLength);

        //Update the data length now that we know the exact size of it (include the UDH length in octets also if necessary)
        if (nSMSMessages > 1)
          nThisMessageLength += 6;
#pragma warning(suppress: 26446 26472)
        SMS[nDataLengthIndex] = static_cast<BYTE>(nThisMessageLength);
        break;
      }
      default:
      {
#pragma warning(suppress: 26477)
        ATLASSERT(FALSE);
        break;
      }
    }

    //Convert the BYTE array representation to the string version
    std::string sData;
    ConvertToHexCharacters(SMS, sData);
#pragma warning(suppress: 26472)
    const int nMsgSize{static_cast<int>((sData.length() / 2) - 1)}; //Always exclude the SMSC data octet

    //Prepare to send the PDU
    ATL::CAtlStringA sCMD;
    sCMD.Format("AT+CMGS=%d\r", nMsgSize);
    Write(sCMD, sCMD.GetLength());

    //And finally write out the PDU as a string
#pragma warning(suppress: 26472)
    Write(sData.c_str(), static_cast<DWORD>(sData.length()));
    Write("\x1A", 1);
    if (!WaitForSerialResponse())
      ThrowSMSException(E_SMS_SEND_MESSAGE, m_sLastResponse);

    //Prepare for the next loop around
#pragma warning(suppress: 26481)
    pbyUserData += nSMSMaxLength;
    nMessageLength -= nSMSMaxLength;
  }
}

void CSMSSerialPort::AppendEncodedSMSCNumber(_In_ const CMFCSMSString& sNumber, _In_ BYTE byTypeOfNumber, _Inout_ std::vector<BYTE>& message)
{
  //Form a ASCII version of the number
  ATL::CT2A sAsciiNumber{sNumber.c_str()};
  const size_t nLength{strlen(sAsciiNumber.operator LPSTR())};
  if (nLength)
  {
    const size_t nOctets{(nLength + 1) / 2};

    //Validate the size
    if ((nOctets + 1) > 255)
      CSMSSerialPort::ThrowSMSException(E_SMS_SMSC_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE);

    message.push_back(static_cast<BYTE>(nOctets + 1)); //The length of the number
    message.push_back(byTypeOfNumber); //The type of address

    //Now the number itself
    for (size_t i{0}; i<nOctets; i++)
    {
#pragma warning(suppress: 26472 26481)
      const BYTE byNibble1{static_cast<BYTE>(sAsciiNumber[i * 2] - '0')};
      BYTE byNibble2{0};
      const size_t nNibble2Offset{(i * 2) + 1};
      if (nNibble2Offset < nLength)
#pragma warning(suppress: 26472 26481)
        byNibble2 = static_cast<BYTE>(sAsciiNumber[nNibble2Offset] - '0');
      else
        byNibble2 = 0xF;

#pragma warning(suppress: 26489)
      message.push_back(static_cast<BYTE>((byNibble2 << 4) | byNibble1));
    }
  }
  else
    message.push_back(0x0);
}

void CSMSSerialPort::AppendEncodedPhoneNumber(_In_ const CMFCSMSString& sNumber, _In_ BYTE byTypeOfNumber, _Inout_ std::vector<BYTE>& message)
{
  //Form a ASCII version of the number
  ATL::CT2A sAsciiNumber{sNumber.c_str()};
  const size_t nLength{strlen(sAsciiNumber.operator LPSTR())};
  if (nLength)
  {
    const size_t nOctets{(nLength + 1) / 2};

    //Validate the size
    if (nLength > 255)
      CSMSSerialPort::ThrowSMSException(E_SMS_PHONE_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE);

    message.push_back(static_cast<BYTE>(nLength)); //The length of the number
    message.push_back(byTypeOfNumber); //The type of address

    //Now the number itself
    for (size_t i{0}; i<nOctets; i++)
    {
#pragma warning(suppress: 26472 26481)
      const BYTE byNibble1{static_cast<BYTE>(sAsciiNumber[i * 2] - '0')};
      BYTE byNibble2{0};
      const size_t nNibble2Offset{(i * 2) + 1};
      if (nNibble2Offset < nLength)
#pragma warning(suppress: 26472 26481)
        byNibble2 = static_cast<BYTE>(sAsciiNumber[nNibble2Offset] - '0');
      else
        byNibble2 = 0xF;
#pragma warning(suppress: 26489)
      message.push_back(static_cast<BYTE>((byNibble2 << 4) | byNibble1));
    }
  }
  else
    message.push_back(0x0);
}

void CSMSSerialPort::AppendGSM7Bit(_In_reads_bytes_opt_(nDataLength) const BYTE* pbyData, _In_ int nDataLength, _In_ int nPaddingBits, _Inout_ std::vector<BYTE>& EncodedData)
{
  int nBits{0};
  if (nPaddingBits && nDataLength)
  {
    nBits = 7 - nPaddingBits;
#pragma warning(suppress: 26477)
    ATLASSUME(pbyData != nullptr);
#pragma warning(suppress: 26481)
    EncodedData.push_back(static_cast<BYTE>(pbyData[0] << (7 - nBits)));
    nBits++;
  }

  for (int i{0}; i<nDataLength; i++)
  {
    if (nBits == 7)
    {
      nBits = 0;
      continue;
    }
#pragma warning(suppress: 26477)
    ATLASSUME(pbyData != nullptr);
#pragma warning(suppress: 26472 26481)
    BYTE byOctet{static_cast<BYTE>((pbyData[i] & 0x7F) >> nBits)}; //NOLINT(clang-analyzer-core.NullDereference)
    if (i < (nDataLength - 1))
    {
#pragma warning(suppress: 6385 26481)
      byOctet |= (pbyData[i + 1] << (7 - nBits));
    }
    EncodedData.push_back(byOctet);
    nBits++;
  }
}

void CSMSSerialPort::ConvertToHexCharacters(_In_ const std::vector<BYTE>& message, _Inout_ std::string& sEncodedData)
{
#pragma warning(suppress: 26472)
  const int nArraySize{static_cast<int>(message.size())};
  const int nBufferSize{(nArraySize * 2) + 1};
  sEncodedData.resize(nBufferSize);
#pragma warning(suppress: 26429 26446)
  char* pszEncodedData{sEncodedData.data()};
  int nPDUOutputIndex{0};
  for (int i{0}; i<nArraySize; i++)
  {
#pragma warning(suppress: 26446)
    const BYTE byData{message[i]};
    const int nNibble1{(byData & 0xF0) >> 4};
    if (nNibble1 > 9)
#pragma warning(suppress: 26472 26481)
      pszEncodedData[nPDUOutputIndex] = static_cast<char>(nNibble1 - 10 + 'A');
    else
#pragma warning(suppress: 26472 26481)
      pszEncodedData[nPDUOutputIndex] = static_cast<char>(nNibble1 + '0');
    ++nPDUOutputIndex;
    const int nNibble2{byData & 0xF};
    if (nNibble2 > 9)
#pragma warning(suppress: 26472 26481)
      pszEncodedData[nPDUOutputIndex] = static_cast<char>(nNibble2 - 10 + 'A');
    else
#pragma warning(suppress: 26472 26481)
      pszEncodedData[nPDUOutputIndex] = static_cast<char>(nNibble2 + '0');
    ++nPDUOutputIndex;
  }
}
