/*
Module : MfcSMS.h
Purpose: Declaration for MFC wrapper classes for sending SMS messages
Created: PJN / 27-10-2002

Copyright (c) 2002 - 2022 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise) 
when your product is released in binary form. You are allowed to modify the source code in any way you want 
except you cannot modify the copyright details at the top of each module. If you want to distribute source 
code with your application, then you are only allowed to distribute versions released by the author. This is 
to maintain a single distribution point for the source code. 

*/


//////////////////// Macros / Defines /////////////////////////////////////////

#pragma once

#ifndef __MFCSMS_H__
#define __MFCSMS_H__

#ifndef CMFCSMS_EXT_CLASS
#define CMFCSMS_EXT_CLASS
#endif //#ifndef CMFCSMS_EXT_CLASS

//The list of currently defined custom error codes
#define E_SMS_FAIL_CREATE_PDU_MESSAGE                     MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_FAIL_CREATE_PDU_MESSAGE)
#define E_SMS_NO_MODEM_RESPONSE                           MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_FAIL_RECEIVE_RESPONSE_FROM_MODEM)
#define E_SMS_NO_SMS_SUPPORT                              MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_MODEM_DOES_NOT_SUPPORT_SMS)
#define E_SMS_SET_PDU_MODE                                MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_FAIL_SET_PDU_MODE)
#define E_SMS_SEND_MESSAGE                                MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_FAIL_SEND_SMS_MESSAGE)
#define E_SMS_SMSC_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE  MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_SMSC_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE)
#define E_SMS_PHONE_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_PHONE_NUMBER_TOO_LARGE_TO_SEMI_OCTET_ENCODE)
#define E_SMS_MESSAGE_TOO_LARGE_TO_PDU_ENCODE             MAKE_HRESULT(SEVERITY_ERROR, FACILITY_ITF, IDS_SMSSERIALPORT_MESSAGE_TOO_LARGE_TO_PDU_ENCODE)

#ifndef __ATLTIME_H__
#pragma message("To avoid this message, please put atltime.h in your pre compiled header (normally stdafx.h)")
#include <atltime.h>
#endif //#ifndef __ATLTIME_H__

#ifndef _STRING_
#pragma message("To avoid this message, please put string in your pre compiled header (normally stdafx.h)")
#include <string>
#endif //#ifndef _STRING_

#ifndef _VECTOR_
#pragma message("To avoid this message, please put vector in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _VECTOR_

#ifndef _ARRAY_
#pragma message("To avoid this message, please put array in your pre compiled header (normally stdafx.h)")
#include <array>
#endif //#ifndef _ARRAY_

#ifndef _EXCEPTION_
#pragma message("To avoid this message, please put exception in your pre compiled header (normally stdafx.h)")
#include <vector>
#endif //#ifndef _EXCEPTION_

#ifdef _UNICODE
using CMFCSMSString = std::wstring;
#else
using CMFCSMSString = std::string;
#endif //#ifdef _UNICODE


//////////////////// Classes //////////////////////////////////////////////////


///////// The exception class for the framework ///////////////////////////////

class CSMSException : public std::exception
{
public:
//Constructors / Destructors
  CSMSException(_In_ HRESULT hr, _In_ CMFCSMSString sLastResponse) noexcept;
  CSMSException(_In_ DWORD dwError, _In_ DWORD dwFacility, _In_ CMFCSMSString sLastResponse) noexcept;
  CSMSException(const CSMSException&) = default;
  CSMSException(CSMSException&&) = default;
  ~CSMSException() = default; //NOLINT(modernize-use-override)

//Methods
  CSMSException& operator=(const CSMSException&) = default;
  CSMSException& operator=(CSMSException&&) = default;
  virtual BOOL GetErrorMessage(_Out_z_cap_(nMaxError) LPTSTR lpszError, _In_ UINT nMaxError, _Out_opt_ PUINT pnHelpContext = nullptr);

//Data members
  HRESULT m_hr;
  CMFCSMSString m_sLastResponse;
#pragma warning(suppress: 26495)
};


///////// Class used to report statistics related to an SMS message ///////////

class CMFCSMS_EXT_CLASS CSMSMessageStats
{
public:
//Constructors / Destructors
  CSMSMessageStats() noexcept;
  CSMSMessageStats(const CSMSMessageStats&) = default;
  CSMSMessageStats(CSMSMessageStats&&) = default;
  ~CSMSMessageStats() = default;

//Methods
  CSMSMessageStats& operator=(const CSMSMessageStats&) = default;
  CSMSMessageStats& operator=(CSMSMessageStats&&) = default;

//Member variables
  int m_nUserDataSizeCharacters; //The size of the complete message in characters (before it is broken down)
  int m_nUserDataSizeBytes; //The size of the complete message in bytes (before it is broken down)
  int m_nMessages; //How many SMS messages this will encode into
  int m_nMaxBytesPerMessage; //The maximum bytes available in each message
  int m_nMaxCharactersPerMessage; //How maximum characters available in each message
  int m_nExtendedCharacters; //How many GSM 7-bit escaped characters are in the message 
  int m_nReplacementCharacters; //How many replacement characters were used
};


///////// Simple class encapsulation of an SMS message ////////////////////////

class CMFCSMS_EXT_CLASS CSMSMessage
{
public:
//Enums
  enum class Alphabet
  {
    GSM_7BIT = 0,
    UCS2 = 1,
    EIGHTBIT = 2
  };

  enum class PhoneFormat
  {
    UnknownFormat = 0x81,
    InternationalFormat = 0x91,
    NationalFormat = 0xA1,
  };

//Constructors / Destructors
  CSMSMessage() noexcept;
  CSMSMessage(const CSMSMessage&) = default;
  CSMSMessage(CSMSMessage&&) = default;
  ~CSMSMessage() = default;

//Methods
  CSMSMessage& operator=(const CSMSMessage&) = default;
  CSMSMessage& operator=(CSMSMessage&&) = default;
  CSMSMessageStats Statistics(); //Generate encoding statistics for the current message

//Static Methods
  static void ConvertToGSM7Bit(_In_ const CMFCSMSString& sData, _In_ TCHAR cReplacementCharacter, _Inout_ std::vector<BYTE>& EncodedData, _Out_opt_ int* pnExtendedCharacters, _Out_opt_ int* pnReplacementCharacters);
  static void ConvertToUCS2(_In_ const CMFCSMSString& sData, _Inout_ std::vector<BYTE>& EncodedData);
  static BYTE TCHAR2BYTE(_In_ TCHAR cCharacter);

//Member variables
  CMFCSMSString m_sMessage; //The message contents 
  std::vector<BYTE> m_byMessage; //The message contents (if being sent as 8 bit)
  CMFCSMSString m_sServiceCenterNumber; //The Service Center Phone number to use. Left empty will use the Phone's built in SMSC
  BYTE m_ServiceCenterNumberAddressType; //The type of address for "m_sServiceCenterNumber". e.g. 0x91 would be international, while 0xA1 would be National
  CMFCSMSString m_sPhoneNumber; //The destination Phone number for this message
  BYTE m_PhoneNumberAddressType; //The type of address for "m_sPhoneNumber"
  bool m_bUseValidityPeriod; //Should a validity period be used with this message.
  bool m_bUseRelativeValidityPeriod; //true if "m_RelativeValidityPeriod" should be used, otherwise use "m_AbsoluteValidityPeriod"
  ATL::CTimeSpan m_RelativeValidityPeriod; //The relative time after which if this message is not delivered, it can be discarded
  SYSTEMTIME m_AbsoluteValidityPeriod; //The absolute time (local time) after which if this message is not delivered, it can be discarded
  long m_TimeZone; //The timezone value (in minutes) associated with m_AbsoluteValidityPeriod. This would be used to 
                   //make conversions to UTC for date calculations
  bool m_bRequireStatusReport; //Should a status report for this message be generated
  TCHAR m_cReplacementCharacter; //The replacement character to use if any data in "m_sMessage" cannot be mapped to a 7 bit SMS character
  Alphabet m_Alphabet; //The encoding to use for sending the message
  BYTE m_byReferenceNumber; //The reference number to use for this message if we end up generating a concatenated SMS
};


//// Class which allows an SMS message to be transmitted via a serial port ////

class CMFCSMS_EXT_CLASS CSMSSerialPort
{
public:
//Enums
  enum
  {
    SMS_SUBMIT = 0x01,
    UseRelativeValidityPeriod = 0x10,
    DoNoUseRelativeValidityPeriod = 0x18,
    RequireStatusReport = 0x20,
    UseUDH = 0x40,
    Empty_TP_Message_Reference = 0x00,
    SMS_Type0_Protocol_Identifier = 0x00,
    Data_Coding_GSM_7Bit = 0x00,
    Data_Coding_8Bit = 0x04,
    Data_Coding_UCS2 = 0x08,
    IEI_ConcatenatedSMS = 0x00
  };

//Constructors / Destructors
  CSMSSerialPort() noexcept;
  CSMSSerialPort(const CSMSSerialPort&) = delete;
  CSMSSerialPort(CSMSSerialPort&&) = delete;
  virtual ~CSMSSerialPort();

//Methods
  CSMSSerialPort& operator=(const CSMSSerialPort&) = delete;
  CSMSSerialPort& operator=(CSMSSerialPort&&) = delete;
  void Send(_In_ const CSMSMessage& message);
  virtual void Open(_In_ int nPort, _In_ DWORD dwTimeout = 5000);
  void Close() noexcept;
  operator HANDLE() const noexcept { return m_hComm; };
  _NODISCARD bool IsOpen() const noexcept { return m_hComm != INVALID_HANDLE_VALUE; };
  _NODISCARD CMFCSMSString GetLastResponse() const { return m_sLastResponse; };
  static void ThrowSMSException(_In_ DWORD dwError = 0, _In_ DWORD Facility = FACILITY_WIN32, _In_ const CMFCSMSString& sLastResponse = CMFCSMSString());
  static void ThrowSMSException(_In_ HRESULT hr, _In_ const CMFCSMSString& sLastResponse = CMFCSMSString());

protected:
//Methods
  virtual DWORD Read(_Out_writes_bytes_(dwCount) __out_data_source(FILE) void* lpBuf, _In_ DWORD dwCount);
  virtual DWORD Write(_In_reads_bytes_opt_(dwCount) const void* lpBuf, _In_ DWORD dwCount);
  virtual void InitializePort(_In_ DWORD dwTimeout);
  virtual bool DataWaiting(_In_ DWORD dwTimeout);
  virtual void GetOverlappedResult(_In_ bool bWait, _Out_ DWORD& dwBytesTransferred);
  virtual bool WaitForSerialResponse();
  static void AppendEncodedSMSCNumber(_In_ const CMFCSMSString& sNumber, _In_ BYTE byTypeOfNumber, _Inout_ std::vector<BYTE>& message);
  static void AppendEncodedPhoneNumber(_In_ const CMFCSMSString& sNumber, _In_ BYTE byTypeOfNumber, _Inout_ std::vector<BYTE>& message);
  static void AppendGSM7Bit(_In_reads_bytes_opt_(nDataLength) const BYTE* pbyData, _In_ int nDataLength, _In_ int nPaddingBits, _Inout_ std::vector<BYTE>& EncodedData);
  static void ConvertToHexCharacters(_In_ const std::vector<BYTE>& message, _Inout_ std::string& sEncodedData);
  constexpr static BYTE SwappedNibble(_In_ int nValue)
  {
#pragma warning(suppress: 26472)
    BYTE byReturn = static_cast<BYTE>((nValue % 10) << 4);
    byReturn |= (nValue / 10);
    return byReturn;
  }

//Member variables
  HANDLE m_hComm;
  HANDLE m_hEvent;
  CMFCSMSString m_sLastResponse;
  DWORD m_dwTimeout;
  OVERLAPPED m_Overlapped;
};

#endif //#ifndef __MFCSMS_H__
